/*
 * $Id: awk.h 259 2009-08-20 11:28:03Z hyunghwan.chung $
 *
   Copyright 2006-2009 Chung, Hyung-Hwan.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

#ifndef _QSE_AWK_AWK_H_
#define _QSE_AWK_AWK_H_

#include <qse/types.h>
#include <qse/macros.h>
#include <qse/cmn/map.h>
#include <qse/cmn/str.h>

/** @file
 * An embeddable AWK interpreter is defined in this header file.
 *
 * @example awk.c
 * This program demonstrates how to build a complete awk interpreter.
 * @example awk01.c
 * This program demonstrates how to use qse_awk_rtx_loop().
 * @example awk02.c
 * The program deparses the source code and prints it before executing it.
 * @example awk03.c
 * This program demonstrates how to use qse_awk_rtx_call().
 * It parses the program stored in the string src and calls the functions
 * stated in the array fnc. If no errors occur, it should print 24.
 * @example awk04.c
 * This programs shows how to qse_awk_rtx_call().
 */

/** @struct qse_awk_t
 * The #qse_awk_t type defines an AWK interpreter. It provides an interface
 * to parse an AWK script and run it to manipulate input and output data.
 *
 * In brief, you need to call APIs with user-defined handlers to run a typical
 * AWK script as shown below:
 * 
 * @code
 * qse_awk_t* awk;
 * qse_awk_rtx_t* rtx;
 * qse_awk_sio_t sio; // need to initialize it with callback functions
 * qse_awk_rio_t rio; // need to initialize it with callback functions
 * qse_cstr_t args[5]; // need to initialize it with strings
 *
 * awk = qse_awk_open (mmgr, 0, prm); // create an interpreter 
 * qse_awk_parse (awk, &sio);          // parse a script 
 * rtx = qse_awk_rtx_open (awk, 0, &rio, args); // create a runtime context 
 * retv = qse_awk_rtx_loop (rtx);     // run a standard AWK loop 
 * if (retv != QSE_NULL) 
 *    qse_awk_rtx_refdownval (rtx, retv); // free return value
 * qse_awk_rtx_close (rtx);           // destroy the runtime context
 * qse_awk_close (awk);               // destroy the interpreter
 * @endcode
 *
 * It provides an interface to change the conventional behavior of the 
 * interpreter; most notably, you can call a particular function with 
 * qse_awk_rtx_call() instead of entering the BEGIN,pattern-action blocks,END
 * loop. By doing this, you may utilize a script in an event-driven way.
 *
 * @sa qse_awk_rtx_t qse_awk_open qse_awk_close
 */
typedef struct qse_awk_t qse_awk_t;

/** @struct qse_awk_rtx_t
 * The #qse_awk_rtx_t type defines a runtime context. A runtime context 
 * maintains runtime state for a running script. You can create multiple
 * runtime contexts out of a single AWK interpreter; in other words, you 
 * can run the same script with different input and output data by providing
 * customized I/O handlers when creating a runtime context with 
 * qse_awk_rtx_open().
 *
 * I/O handlers are categoriezed into three kinds: console, file, pipe.
 * The #qse_awk_rio_t type defines as a callback a set of I/O handlers 
 * to handle runtime I/O:
 * - getline piped in from a command reads from a pipe.
 *   ("ls -l" | getline line)
 * - print and printf piped out to a command writes to a pipe.
 *   (print 2 | "sort")
 * - getline redirected in reads from a file.
 *   (getline line < "file")
 * - print and printf redirected out writes to a file.
 *   (print num > "file")
 * - The pattern-action loop and getline with no redirected input
 *   reads from a console. (/susie/ { ... })
 * - print and printf writes to a console. (print "hello, world")
 *
 * @sa qse_awk_t qse_awk_rtx_open qse_awk_rio_t
 */
typedef struct qse_awk_rtx_t qse_awk_rtx_t;

/**
 * The QSE_AWK_VAL_HDR defines the common header for a value.
 * Three common fields are:
 * - type - type of a value from #qse_awk_val_type_t
 * - ref - reference count
 * - nstr - numeric string marker
 */
#if QSE_SIZEOF_INT == 2
#	define QSE_AWK_VAL_HDR \
		unsigned int type: 3; \
		unsigned int ref: 11; \
		unsigned int nstr: 2
#else
#	define QSE_AWK_VAL_HDR \
		unsigned int type: 3; \
		unsigned int ref: 27; \
		unsigned int nstr: 2
#endif

/**
 * The qse_awk_val_t type is an abstract value type. A value commonly contains:
 * - type of a value
 * - reference count
 * - indicator for a numeric string
 */
struct qse_awk_val_t
{
	QSE_AWK_VAL_HDR;	
};
typedef struct qse_awk_val_t qse_awk_val_t;

/**
 * The qse_awk_val_nil_t type is a nil value type. The type field is 
 * #QSE_AWK_VAL_NIL.
 */
struct qse_awk_val_nil_t
{
	QSE_AWK_VAL_HDR;
};
typedef struct qse_awk_val_nil_t  qse_awk_val_nil_t;

/**
 * The qse_awk_val_int_t type is an integer number type. The type field is
 * #QSE_AWK_VAL_INT.
 */
struct qse_awk_val_int_t
{
	QSE_AWK_VAL_HDR;
	qse_long_t val;
	void*      nde;
};
typedef struct qse_awk_val_int_t qse_awk_val_int_t;

/**
 * The qse_awk_val_real_t type is a floating-point number type. The type field
 * is #QSE_AWK_VAL_REAL.
 */
struct qse_awk_val_real_t
{
	QSE_AWK_VAL_HDR;
	qse_real_t val;
	void*      nde;
};
typedef struct qse_awk_val_real_t qse_awk_val_real_t;

/**
 * The qse_awk_val_str_t type is a string type. The type field is
 * #QSE_AWK_VAL_STR.
 */
struct qse_awk_val_str_t
{
	QSE_AWK_VAL_HDR;
	qse_char_t* ptr;
	qse_size_t  len;
};
typedef struct qse_awk_val_str_t  qse_awk_val_str_t;

/**
 * The qse_awk_val_rex_t type is a regular expression type.  The type field 
 * is #QSE_AWK_VAL_REX.
 */
struct qse_awk_val_rex_t
{
	QSE_AWK_VAL_HDR;
	qse_char_t* ptr;
	qse_size_t  len;
	void*       code;
};
typedef struct qse_awk_val_rex_t  qse_awk_val_rex_t;

/**
 * The qse_awk_val_map_t type defines a map type. The type field is 
 * #QSE_AWK_VAL_MAP.
 */
struct qse_awk_val_map_t
{
	QSE_AWK_VAL_HDR;

	/* TODO: make val_map to array if the indices used are all 
	 *       integers switch to map dynamically once the 
	 *       non-integral index is seen.
	 */
	qse_map_t* map; 
};
typedef struct qse_awk_val_map_t  qse_awk_val_map_t;

/**
 * The qse_awk_val_ref_t type defines a reference type that is used
 * internally only. The type field is #QSE_AWK_VAL_REF.
 */
struct qse_awk_val_ref_t
{
	QSE_AWK_VAL_HDR;

	enum
	{
		/* keep these items in the same order as corresponding items
		 * in tree.h */
		QSE_AWK_VAL_REF_NAMED,    /**< plain named variable */
		QSE_AWK_VAL_REF_GBL,      /**< plain global variable */
		QSE_AWK_VAL_REF_LCL,      /**< plain local variable */
		QSE_AWK_VAL_REF_ARG,      /**< plain function argument */
		QSE_AWK_VAL_REF_NAMEDIDX, /**< member of named map variable */
		QSE_AWK_VAL_REF_GBLIDX,   /**< member of global map variable */
		QSE_AWK_VAL_REF_LCLIDX,   /**< member of local map variable */
		QSE_AWK_VAL_REF_ARGIDX,   /**< member of map argument */
		QSE_AWK_VAL_REF_POS       /**< positional variable */
	} id;

	/* if id is QSE_AWK_VAL_REF_POS, adr holds an index of the 
	 * positional variable. Otherwise, adr points to the value 
	 * directly. */
	qse_awk_val_t** adr;
};
typedef struct qse_awk_val_ref_t  qse_awk_val_ref_t;

typedef qse_real_t (*qse_awk_pow_t) (
	qse_awk_t* awk,
	qse_real_t x, 
	qse_real_t y
);

typedef int (*qse_awk_sprintf_t) (
	qse_awk_t*        awk,
	qse_char_t*       buf,
	qse_size_t        size, 
	const qse_char_t* fmt,
	...
);

/**
 * The qse_awk_fnc_fun_t type defines a intrinsic function handler.
 */
typedef int (*qse_awk_fnc_fun_t) (
	qse_awk_rtx_t*    rtx,  /**< runtime context */
	const qse_cstr_t* name  /**< function name */
);

/**
 * The qse_awk_sio_cmd_t type defines I/O commands for a script stream.
 */
enum qse_awk_sio_cmd_t
{
	QSE_AWK_SIO_OPEN   = 0, /**< open a script stream */
	QSE_AWK_SIO_CLOSE  = 1, /**< close a script stream */
	QSE_AWK_SIO_READ   = 2, /**< read text from an input script stream */
	QSE_AWK_SIO_WRITE  = 3  /**< write text to an output script stream */
};
typedef enum qse_awk_sio_cmd_t qse_awk_sio_cmd_t;

typedef struct qse_awk_sio_lxc_t qse_awk_sio_lxc_t;
struct qse_awk_sio_lxc_t
{
	qse_cint_t        c;
	qse_size_t        lin;
	qse_size_t        col;
	const qse_char_t* file;
};

struct qse_awk_sio_arg_t 
{
	const qse_char_t* name;   /**< [IN] name of I/O object */
	void* handle;             /**< [OUT] I/O handle set by a handler */

	/*-- from here down, internal use only --*/
	struct
	{
		qse_char_t buf[1024];
		qse_size_t pos;
		qse_size_t len;
	} b;

	qse_size_t lin;
	qse_size_t col;

	qse_awk_sio_lxc_t last;
	struct qse_awk_sio_arg_t* next;
};
typedef struct qse_awk_sio_arg_t qse_awk_sio_arg_t;

/**
 * The qse_awk_sio_fun_t type defines a source IO function
 */
typedef qse_ssize_t (*qse_awk_sio_fun_t) (
	qse_awk_t*         awk,
	qse_awk_sio_cmd_t  cmd, 
	qse_awk_sio_arg_t* arg,
	qse_char_t*        data,
	qse_size_t         count
);

/**
 * The qse_awk_rio_cmd_t type defines runtime I/O request types.
 */
enum qse_awk_rio_cmd_t
{
	QSE_AWK_RIO_OPEN   = 0, /**< open a stream */
	QSE_AWK_RIO_CLOSE  = 1, /**< close a stream */
	QSE_AWK_RIO_READ   = 2, /**< read a stream */
	QSE_AWK_RIO_WRITE  = 3, /**< write a stream */
	QSE_AWK_RIO_FLUSH  = 4, /**< write buffered data to a stream */
	QSE_AWK_RIO_NEXT   = 5  /**< close the current stream and 
	                             open the next stream (only for console) */
};
typedef enum qse_awk_rio_cmd_t qse_awk_rio_cmd_t;

/**
 * The qse_awk_rio_mode_t type defines the I/O modes used by I/O handlers.
 * Each I/O handler should inspect the requested mode and open an I/O
 * stream accordingly for subsequent operations.
 */
enum qse_awk_rio_mode_t
{
	QSE_AWK_RIO_PIPE_READ      = 0, /**< open a pipe for read */
	QSE_AWK_RIO_PIPE_WRITE     = 1, /**< open a pipe for write */
	QSE_AWK_RIO_PIPE_RW        = 2, /**< open a pipe for read and write */

	QSE_AWK_RIO_FILE_READ      = 0, /**< open a file for read */
	QSE_AWK_RIO_FILE_WRITE     = 1, /**< open a file for write */
	QSE_AWK_RIO_FILE_APPEND    = 2, /**< open a file for append */

	QSE_AWK_RIO_CONSOLE_READ   = 0, /**< open a console for read */
	QSE_AWK_RIO_CONSOLE_WRITE  = 1  /**< open a console for write */
};
typedef enum qse_awk_rio_mode_t qse_awk_rio_mode_t;

/**
 * The qse_awk_rio_arg_t defines the data structure passed to a runtime 
 * I/O handler. An I/O handler should inspect the @a mode field and the 
 * @a name field and store an open handle to the @handle field when 
 * #QSE_AWK_RIO_OPEN is requested. For other request type, it can refer
 * to the handle field set previously.
 */
struct qse_awk_rio_arg_t 
{
	qse_awk_rio_mode_t mode;  /**< [IN] I/O mode */
	qse_char_t* name;         /**< [IN] name of I/O object */
	void* handle;             /**< [OUT] I/O handle set by a handler */

	/*--  from here down, internal use only --*/
	int type; 

	struct
	{
		qse_char_t buf[2048];
		qse_size_t pos;
		qse_size_t len;
		qse_bool_t eof;
		qse_bool_t eos;
	} in;

	struct
	{
		qse_bool_t eof;
		qse_bool_t eos;
	} out;

	struct qse_awk_rio_arg_t* next;
};
typedef struct qse_awk_rio_arg_t qse_awk_rio_arg_t;

/**
 * The qse_awk_rio_fun_t type defines a runtime I/O handler.
 */
typedef qse_ssize_t (*qse_awk_rio_fun_t) (
	qse_awk_rtx_t*      rtx,
	qse_awk_rio_cmd_t   cmd,
	qse_awk_rio_arg_t*  riod,
	qse_char_t*         data,
	qse_size_t          count
);

/**
 * The qse_awk_prm_t type defines primitive functions
 */
struct qse_awk_prm_t
{
	qse_awk_pow_t     pow;
	qse_awk_sprintf_t sprintf;

#if 0
	/* TODO: accept regular expression handling functions */
	void* (*build) (
		qse_awk_t*        awk,
		const qse_char_t* ptn, 
		qse_size_t        len, 
		int*              errnum
	);

	int (*match) (
		qse_awk_t*         awk,
		void*              code,
		int                option,
		const qse_char_t*  str,
		qse_size_t         len, 
		const qse_char_t** mptr,
		qse_size_t*        mlen, 
		int*               errnum
	);

	void (*free) (
		qse_awk_t* awk,
		void*      code
	);

	qse_bool_t (*isempty) (
		qse_awk_t* awk,
		void*      code
	);
#endif
};
typedef struct qse_awk_prm_t qse_awk_prm_t;

/**
 * The qse_awk_sio_t type defines a script stream handler set.
 * The qse_awk_parse() function calls the input and output handler to parse
 * a script and optionally deparse it. Typical input and output handlers 
 * are shown below:
 *
 * @code
 * qse_ssize_t in (
 *    qse_awk_t* awk, qse_awk_sio_cmd_t cmd,
 *    qse_char_t* buf, qse_size_t size)
 * {
 *    if (cmd == QSE_AWK_SIO_OPEN) open input stream;
 *    else if (cmd == QSE_AWK_SIO_CLOSE) close input stream;
 *    else read input stream and fill buf up to size characters;
 * }
 *
 * qse_ssize_t out (
 *    qse_awk_t* awk, qse_awk_sio_cmd_t cmd,
 *    qse_char_t* data, qse_size_t size)
 * {
 *    if (cmd == QSE_AWK_SIO_OPEN) open_output_stream;
 *    else if (cmd == QSE_AWK_SIO_CLOSE) close_output_stream;
 *    else write data of size characters to output stream;
 * }
 * @endcode
 *
 * For #QSE_AWK_SIO_OPEN, a handler must return:
 * - -1 if it failed to open a stream.
 * - 0 if it has opened a stream but has reached the end.
 * - 1 if it has successfully opened a stream.
 * 
 * For #QSE_AWK_SIO_CLOSE, a handler must return:
 * - -1 if it failed to close a stream.
 * - 0 if it has closed a stream.
 *
 * For #QSE_AWK_SIO_READ and #QSE_AWK_SIO_WRITE, a handler must return:
 * - -1 if there was an error occurred during operation.
 * - 0 if it has reached the end.
 * - the number of characters read or written on success.
 */
struct qse_awk_sio_t
{
	qse_awk_sio_fun_t in;  /**< input script stream handler */
	qse_awk_sio_fun_t out; /**< output script stream handler */
};
typedef struct qse_awk_sio_t qse_awk_sio_t;

/**
 * The qse_awk_rio_t type defines a runtime I/O handler set.
 * @sa qse_awk_rtx_t
 */
struct qse_awk_rio_t
{
	qse_awk_rio_fun_t pipe;    /**< pipe handler */
	qse_awk_rio_fun_t file;    /**< file handler */
	qse_awk_rio_fun_t console; /**< console handler */
};
typedef struct qse_awk_rio_t qse_awk_rio_t;

/**
 * The qse_awk_rcb_t type defines runtime callbacks. You may set callbacks
 * with qse_awk_rtx_setrcb() to be informed of important events during runtime.
 */
struct qse_awk_rcb_t
{
	/**
	 * called by qse_awk_rtx_loop() and qse_awk_rtx_call() for
	 * each statement executed.
	 */
	void (*on_statement) (
		qse_awk_rtx_t* rtx, qse_size_t line, void* udd);

	/**
	 * A caller may store a custom data pointer into this field.
	 */
	void* udd;
};
typedef struct qse_awk_rcb_t qse_awk_rcb_t;

/**
 * The qse_awk_option_t type defines various options to change the behavior
 * of #qse_awk_t.
 */
enum qse_awk_option_t
{ 
	/**
	 * allows undeclared variables and implicit concatenation 
	 **/
	QSE_AWK_IMPLICIT    = (1 << 0),

	/** 
	 * allows explicit variable declaration, the concatenation
	 * operator, a period, and performs the parse-time function check. 
	 */
	QSE_AWK_EXPLICIT    = (1 << 1), 

	/** 
	 * supports extra operators:
	 * - @b <<, <<= left-shift
	 * - @b >>, >>= right-shiftt
	 * - @b ^^, ^^= xor
	 * - @b ~  bitwise-not
	 * - @b // idiv (get quotient)
	 */
	QSE_AWK_EXTRAOPS    = (1 << 2), 

	/** supports @b getline and @b print */
	QSE_AWK_RIO         = (1 << 3), 

	/** supports dual direction pipe if #QSE_AWK_RIO is on */
	QSE_AWK_RWPIPE      = (1 << 4),

	/** a new line can terminate a statement */
	QSE_AWK_NEWLINE     = (1 << 5),

	/** 
	 * strips off leading and trailing spaces when splitting a record
	 * into fields with a regular expression. 
	 *
	 * @code
	 * BEGIN { FS="[:[:space:]]+"; } 
	 * { 
	 *    print "NF=" NF; 
	 *    for (i = 0; i < NF; i++) print i " [" $(i+1) "]";
	 * }
	 * @endcode
	 * " a b c " is split to [a], [b], [c] if #QSE_AWK_STRIPRSPC is on.
	 * Otherwise, it is split to [], [a], [b], [c], [].
	 */
	QSE_AWK_STRIPRECSPC    = (1 << 6),

	/**
	 * strips off leading spaces when converting a string to a number.
	 */
	QSE_AWK_STRIPSTRSPC    = (1 << 7),

	/** enables @b nextofile */
	QSE_AWK_NEXTOFILE   = (1 << 8),

	/** enables @b reset */
	QSE_AWK_RESET       = (1 << 9),

	/** CR + LF by default */
	QSE_AWK_CRLF        = (1 << 10),

	/** allows the assignment of a map value to a variable */
	QSE_AWK_MAPTOVAR    = (1 << 11),

	/** allows @b BEGIN, @b END, pattern-action blocks */
	QSE_AWK_PABLOCK     = (1 << 12),

	/** allows {n,m} in a regular expression. */
	QSE_AWK_REXBOUND    = (1 << 13),

	/** 
	 * performs numeric comparison when a string convertable
	 * to a number is compared with a number or vice versa.
	 *
	 * For an expression (9 > "10.9"),
	 * - 9 is greater if #QSE_AWK_NCMPONSTR is off;
	 * - "10.9" is greater if #QSE_AWK_NCMPONSTR is on
	 */
	QSE_AWK_NCMPONSTR = (1 << 14),

	/**
	 * enables the strict naming rule.
	 * - a parameter name can not be the same as the owning function name.
	 * - a local variable name can not be the same as the owning 
	 *   function name.
	 */
	QSE_AWK_STRICTNAMING = (1 << 15),

	/**
	 * supports file inclusion by enabling a keyword 'include'
	 */
	QSE_AWK_INCLUDE = (1 << 16),

	/** 
	 * makes #qse_awk_t to behave as compatibly as classical AWK
	 * implementations 
	 */
	QSE_AWK_CLASSIC  = QSE_AWK_IMPLICIT | QSE_AWK_RIO | 
	                   QSE_AWK_NEWLINE | QSE_AWK_PABLOCK | 
	                   QSE_AWK_STRIPSTRSPC | QSE_AWK_STRICTNAMING
};

/**
 * The qse_awk_errnum_t type defines error codes.
 */
enum qse_awk_errnum_t
{
	QSE_AWK_ENOERR,  /**< no error */
	QSE_AWK_EUNKNOWN,/**< unknown error */

	/* common errors */
	QSE_AWK_EINVAL,  /**< invalid parameter or data */
	QSE_AWK_ENOMEM,  /**< insufficient memory */
	QSE_AWK_ENOSUP,  /**< not supported */
	QSE_AWK_ENOPER,  /**< operation not allowed */
	QSE_AWK_ENOENT,  /**< '${0}' not found */
	QSE_AWK_EEXIST,  /**< '${0}' already exists */
	QSE_AWK_EIOERR,  /**< I/O error */

	/* mostly parse errors */
	QSE_AWK_EOPEN,   /**< cannot open '${0}' */
	QSE_AWK_EREAD,   /**< cannot read '${0}' */
	QSE_AWK_EWRITE,  /**< cannot write '${0}' */
	QSE_AWK_ECLOSE,  /**< cannot close '${0}' */

	QSE_AWK_EINTERN, /**< internal error */
	QSE_AWK_ERUNTIME,/**< general run-time error */
	QSE_AWK_EBLKNST, /**< block nested too deeply */
	QSE_AWK_EEXPRNST,/**< expression nested too deeply */

	QSE_AWK_ELXCHR,  /**< invalid character '${0}' */
	QSE_AWK_ELXDIG,  /**< invalid digit '${0}' */

	QSE_AWK_EEOF,    /**< unexpected end of source */
	QSE_AWK_ECMTNC,  /**< comment not closed properly */
	QSE_AWK_ESTRNC,  /**< string or regular expression not closed */
	QSE_AWK_ELBRACE, /**< left brace expected in place of '${0}' */
	QSE_AWK_ELPAREN, /**< left parenthesis expected in place of '${0}' */
	QSE_AWK_ERPAREN, /**< right parenthesis expected in place of '${0}' */
	QSE_AWK_ERBRACK, /**< right bracket expected in place of '${0}' */
	QSE_AWK_ECOMMA,  /**< comma expected in place of '${0}' */
	QSE_AWK_ESCOLON, /**< semicolon expected in place of '${0}' */
	QSE_AWK_ECOLON,  /**< colon expected in place of '${0}' */
	QSE_AWK_ESTMEND, /**< statement not ending with a semicolon */
	QSE_AWK_EKWIN,   /**< keyword 'in' expected in place of '${0}' */
	QSE_AWK_ENOTVAR, /**< right-hand side of 'in' not a variable */
	QSE_AWK_EEXPRNR, /**< expression not recognized around '${0}' */

	QSE_AWK_EKWFNC,    /**< keyword 'function' expected in place of '${0}' */
	QSE_AWK_EKWWHL,    /**< keyword 'while' expected in place of '${0}' */
	QSE_AWK_EASSIGN,   /**< assignment statement expected */
	QSE_AWK_EIDENT,    /**< identifier expected in place of '${0}' */
	QSE_AWK_EFUNNAME,  /**< '${0}' not a valid function name */
	QSE_AWK_EBLKBEG,   /**< BEGIN not followed by left bracket on the same line */
	QSE_AWK_EBLKEND,   /**< END not followed by left bracket on the same line */
	QSE_AWK_EKWRED,    /**< keyword '${0}' redefined */
	QSE_AWK_EFNCRED,   /**< intrinsic function '${0}' redefined */
	QSE_AWK_EFUNRED,   /**< function '${0}' redefined */
	QSE_AWK_EGBLRED,   /**< global variable '${0}' redefined */
	QSE_AWK_EPARRED,   /**< parameter '${0}' redefined */
	QSE_AWK_EVARRED,   /**< variable '${0}' redefined */
	QSE_AWK_EDUPPAR,   /**< duplicate parameter name '${0}' */
	QSE_AWK_EDUPGBL,   /**< duplicate global variable name '${0}' */
	QSE_AWK_EDUPLCL,   /**< duplicate local variable name '${0}' */
	QSE_AWK_EBADPAR,   /**< '${0}' not a valid parameter name */
	QSE_AWK_EBADVAR,   /**< '${0}' not a valid variable name */
	QSE_AWK_EVARMS,    /**< variable name missing */
	QSE_AWK_EUNDEF,    /**< undefined identifier '${0}' */
	QSE_AWK_ELVALUE,   /**< l-value required */
	QSE_AWK_EGBLTM,    /**< too many global variables */
	QSE_AWK_ELCLTM,    /**< too many local variables */
	QSE_AWK_EPARTM,    /**< too many parameters */
	QSE_AWK_EDELETE,   /**< 'delete' not followed by variable */
	QSE_AWK_ERESET,    /**< 'reset' not followed by variable */
	QSE_AWK_EBREAK,    /**< 'break' outside a loop */
	QSE_AWK_ECONTINUE, /**< 'continue' outside a loop */
	QSE_AWK_ENEXTBEG,  /**< 'next' illegal in BEGIN block */
	QSE_AWK_ENEXTEND,  /**< 'next' illegal in END block */
	QSE_AWK_ENEXTFBEG, /**< 'nextfile' illegal in BEGIN block */
	QSE_AWK_ENEXTFEND, /**< 'nextfile' illegal in END block */
	QSE_AWK_EPRINTFARG,/**< 'printf' not followed by argument */
	QSE_AWK_EPREPST,   /**< both prefix and postfix incr/decr operator present */
	QSE_AWK_EINCDECOPR,/**< illegal operand for incr/decr operator */
	QSE_AWK_EINCLSTR,  /**< 'include' not followed by a string */
	QSE_AWK_EINCLTD,   /**< include level too deep */
	QSE_AWK_EDIRECNR,  /**< directive '${0}' not recognized */

	/* run time error */
	QSE_AWK_EDIVBY0,       /**< divide by zero */
	QSE_AWK_EOPERAND,      /**< invalid operand */
	QSE_AWK_EPOSIDX,       /**< wrong position index */
	QSE_AWK_EARGTF,        /**< too few arguments */
	QSE_AWK_EARGTM,        /**< too many arguments */
	QSE_AWK_EFUNNF,        /**< function '${0}' not found */
	QSE_AWK_ENOTIDX,       /**< variable not indexable */
	QSE_AWK_ENOTDEL,       /**< variable '${0}' not deletable */
	QSE_AWK_ENOTMAP,       /**< value not a map */
	QSE_AWK_ENOTMAPIN,     /**< right-hand side of 'in' not a map */
	QSE_AWK_ENOTMAPNILIN,  /**< right-hand side of 'in' not a map nor nil */
	QSE_AWK_ENOTREF,       /**< value not referenceable */
	QSE_AWK_ENOTASS,       /**< value not assignable */
	QSE_AWK_EIDXVALASSMAP, /**< an indexed value cannot be assigned a map */
	QSE_AWK_EPOSVALASSMAP, /**< a positional cannot be assigned a map */
	QSE_AWK_EMAPTOSCALAR,  /**< map '${0}' not assignable with a scalar */
	QSE_AWK_ESCALARTOMAP,  /**< cannot change a scalar value to a map */
	QSE_AWK_EMAPNA,        /**< map not allowed */
	QSE_AWK_EVALTYPE,      /**< invalid value type */
	QSE_AWK_ERDELETE,      /**< 'delete' called with wrong target */
	QSE_AWK_ERRESET,       /**< 'reset' called with wrong target */
	QSE_AWK_ERNEXTBEG,     /**< 'next' called from BEGIN block */
	QSE_AWK_ERNEXTEND,     /**< 'next' called from END block */
	QSE_AWK_ERNEXTFBEG,    /**< 'nextfile' called from BEGIN block */
	QSE_AWK_ERNEXTFEND,    /**< 'nextfile' called from END block */
	QSE_AWK_EFNCIMPL,      /**< intrinsic function handler for '${0}' failed */
	QSE_AWK_EIOUSER,       /**< wrong user io handler implementation */
	QSE_AWK_EIOIMPL,       /**< I/O callback returned an error */
	QSE_AWK_EIONMNF,       /**< no such I/O name found */
	QSE_AWK_EIONMEM,       /**< I/O name empty */
	QSE_AWK_EIONMNL,       /**< I/O name '${0}' containing '\\0' */
	QSE_AWK_EFMTARG,       /**< not sufficient arguments to formatting sequence */
	QSE_AWK_EFMTCNV,       /**< recursion detected in format conversion */
	QSE_AWK_ECONVFMTCHR,   /**< invalid character in CONVFMT */
	QSE_AWK_EOFMTCHR,      /**< invalid character in OFMT */

	/* regular expression error */
	QSE_AWK_EREXRECUR,     /**< recursion too deep */
	QSE_AWK_EREXRPAREN,    /**< a right parenthesis is expected */
	QSE_AWK_EREXRBRACKET,  /**< a right bracket is expected */
	QSE_AWK_EREXRBRACE,    /**< a right brace is expected */
	QSE_AWK_EREXUNBALPAREN,/**< unbalanced parenthesis */
	QSE_AWK_EREXINVALBRACE,/**< invalid brace */
	QSE_AWK_EREXCOLON,     /**< a colon is expected */
	QSE_AWK_EREXCRANGE,    /**< invalid character range */
	QSE_AWK_EREXCCLASS,    /**< invalid character class */
	QSE_AWK_EREXBRANGE,    /**< invalid boundary range */
	QSE_AWK_EREXEND,       /**< unexpected end of the pattern */
	QSE_AWK_EREXGARBAGE,   /**< garbage after the pattern */

	/* the number of error numbers, internal use only */
	QSE_AWK_NUMERRNUM 
};
typedef enum qse_awk_errnum_t qse_awk_errnum_t;


/** 
 * The qse_awk_errinf_t type defines a placeholder for error information.
 */
struct qse_awk_errinf_t
{
	qse_awk_errnum_t num;      /**< error number */
	qse_char_t       msg[256]; /**< error message */
	qse_size_t       lin;      /**< line number where an error occurred */
};
typedef struct qse_awk_errinf_t qse_awk_errinf_t;

/**
 * The qse_awk_errstr_t type defines an error string getter. It should return 
 * an error formatting string for an error number requested. A new string
 * should contain the same number of positional parameters (${X}) as in the
 * default error formatting string. You can set a new getter into an awk
 * object with the qse_awk_seterrstr() function to customize an error string.
 */
typedef const qse_char_t* (*qse_awk_errstr_t) (
	qse_awk_t*       awk,   /**< awk object */
	qse_awk_errnum_t num    /**< error number */
);

/* depth types */
enum qse_awk_depth_t
{
	QSE_AWK_DEPTH_BLOCK_PARSE = (1 << 0),
	QSE_AWK_DEPTH_BLOCK_RUN   = (1 << 1),
	QSE_AWK_DEPTH_EXPR_PARSE  = (1 << 2),
	QSE_AWK_DEPTH_EXPR_RUN    = (1 << 3),
	QSE_AWK_DEPTH_REX_BUILD   = (1 << 4),
	QSE_AWK_DEPTH_REX_MATCH   = (1 << 5),
	QSE_AWK_DEPTH_INCLUDE     = (1 << 6)
};

/**
 * The qse_awk_gbl_id_t type defines intrinsic globals variable IDs.
 */
enum qse_awk_gbl_id_t
{
	/* this table should match gtab in parse.c.
	 * in addition, qse_awk_rtx_setgbl also counts 
	 * on the order of these values */

	QSE_AWK_GBL_ARGC,
	QSE_AWK_GBL_ARGV,
	QSE_AWK_GBL_CONVFMT,
	QSE_AWK_GBL_FILENAME,
	QSE_AWK_GBL_FNR,
	QSE_AWK_GBL_FS,
	QSE_AWK_GBL_IGNORECASE,
	QSE_AWK_GBL_NF,
	QSE_AWK_GBL_NR,
	QSE_AWK_GBL_OFILENAME,
	QSE_AWK_GBL_OFMT,
	QSE_AWK_GBL_OFS,
	QSE_AWK_GBL_ORS,
	QSE_AWK_GBL_RLENGTH,
	QSE_AWK_GBL_RS,
	QSE_AWK_GBL_RSTART,
	QSE_AWK_GBL_SUBSEP,

	/* these are not not the actual IDs and are used internally only 
	 * Make sure you update these values properly if you add more 
	 * ID definitions, however */
	QSE_AWK_MIN_GBL_ID = QSE_AWK_GBL_ARGC,
	QSE_AWK_MAX_GBL_ID = QSE_AWK_GBL_SUBSEP
};

/**
 * The qse_awk_val_type_t type defines types of AWK values. Each value 
 * allocated is tagged with a value type in the @a type field.
 * @sa qse_awk_val_t QSE_AWK_VAL_HDR
 */
enum qse_awk_val_type_t
{
	/* the values between QSE_AWK_VAL_NIL and QSE_AWK_VAL_STR inclusive
	 * must be synchronized with an internal table of the __cmp_val 
	 * function in run.c */
	QSE_AWK_VAL_NIL  = 0, /**< nil */
	QSE_AWK_VAL_INT  = 1, /**< integer */
	QSE_AWK_VAL_REAL = 2, /**< floating-pointer number */
	QSE_AWK_VAL_STR  = 3, /**< string */

	QSE_AWK_VAL_REX  = 4, /**< regular expression */
	QSE_AWK_VAL_MAP  = 5, /**< map */
	QSE_AWK_VAL_REF  = 6  /**< reference to other types */
};

/**
 * The values defined are used to set the type field of the 
 * qse_awk_rtx_valtostr_out_t structure. The field should be one of the 
 * following values:
 *
 * - QSE_AWK_RTX_VALTOSTR_CPL
 * - QSE_AWK_RTX_VALTOSTR_CPLDUP
 * - QSE_AWK_RTX_VALTOSTR_STRP
 * - QSE_AWK_RTX_VALTOSTR_STRPCAT
 *
 * and it can optionally be ORed with QSE_AWK_RTX_VALTOSTR_PRINT.
 */
enum qse_awk_rtx_valtostr_type_t
{
	QSE_AWK_RTX_VALTOSTR_CPL       = 0x00,
	QSE_AWK_RTX_VALTOSTR_CPLDUP    = 0x01,
	QSE_AWK_RTX_VALTOSTR_STRP      = 0x02,
	QSE_AWK_RTX_VALTOSTR_STRPCAT   = 0x03,
	QSE_AWK_RTX_VALTOSTR_PRINT     = 0x10
};

/**
 * The qse_awk_rtx_valtostr() function converts a value to a string as 
 * indicated in a parameter of the qse_awk_rtx_valtostr_out_t type.
 */
struct qse_awk_rtx_valtostr_out_t
{
	int type;

	union
	{
		qse_xstr_t  cpl;
		qse_xstr_t  cpldup;  /* need to free cpldup.ptr */
		qse_str_t*  strp;
		qse_str_t*  strpcat;
	} u;
};
typedef struct qse_awk_rtx_valtostr_out_t qse_awk_rtx_valtostr_out_t;

#ifdef __cplusplus
extern "C" {
#endif

QSE_DEFINE_COMMON_FUNCTIONS (awk)

/** represents a nil value */
extern qse_awk_val_t* qse_awk_val_nil;

/** represents an empty string  */
extern qse_awk_val_t* qse_awk_val_zls;

/** represents a numeric value -1 */
extern qse_awk_val_t* qse_awk_val_negone;

/** represents a numeric value 0 */
extern qse_awk_val_t* qse_awk_val_zero;

/** represents a numeric value 1 */
extern qse_awk_val_t* qse_awk_val_one;

/**
 * The qse_awk_open() function creates a new qse_awk_t object. The object 
 * created can be passed to other qse_awk_xxx() functions and is valid until 
 * it is destroyed with the qse_awk_close() function. The function saves the 
 * memory manager pointer while it copies the contents of the primitive 
 * function structures. Therefore, you should keep the memory manager valid 
 * during the whole life cycle of an qse_awk_t object.
 *
 * @code
 * qse_awk_t* dummy()
 * {
 *     qse_mmgr_t mmgr;
 *     qse_awk_prm_t prm;
 *     return qse_awk_open (
 *        &mmgr, // NOT OK because the contents of mmgr is 
 *               // invalidated when dummy() returns. 
 *        0, 
 *        &prm   // OK 
 *     );
 * }
 * @endcode
 *
 * @return a pointer to a qse_awk_t object on success, #QSE_NULL on failure.
 */
qse_awk_t* qse_awk_open ( 
	qse_mmgr_t*    mmgr, /**< a memory manager */
	qse_size_t     xtn,  /**< extension size in bytes */
	qse_awk_prm_t* prm   /**< a pointer to a primitive function structure */
);

/**
 *  The qse_awk_close() function destroys a qse_awk_t object.
 * @return 0 on success, -1 on failure 
 */
int qse_awk_close (
	qse_awk_t* awk /**< awk object */
);

/**
 * The qse_awk_getprm() function gets primitive functions
 */
qse_awk_prm_t* qse_awk_getprm (
	qse_awk_t* awk
);

/**
 * The qse_awk_clear() clears the internal state of @a awk. If you want to
 * reuse a qse_awk_t instance that finished being used, you may call 
 * qse_awk_clear() instead of destroying and creating a new
 * #qse_awk_t instance using qse_awk_close() and qse_awk_open().
 *
 * @return 0 on success, -1 on failure
 */
int qse_awk_clear (
	qse_awk_t* awk 
);

/**
 * The qse_awk_geterrstr() gets an error string getter.
 */
qse_awk_errstr_t qse_awk_geterrstr (
	qse_awk_t*       awk    /**< awk object */
);

/**
 * The qse_awk_seterrstr() sets an error string getter that is called to
 * compose an error message when its retrieval is requested.
 *
 * Here is an example of changing the formatting string for the #QSE_SED_ECMDNR 
 * error.
 * @code
 * qse_awk_errstr_t orgerrstr;
 *
 * const qse_char_t* myerrstr (qse_awk_t* awk, qse_awk_errnum_t num)
 * {
 *   if (num == QSE_SED_ECMDNR) return QSE_T("unrecognized command ${0}");
 *   return orgerrstr (awk, num);
 * }
 * int main ()
 * {
 *    qse_awk_t* awk;
 *    ...
 *    orgerrstr = qse_awk_geterrstr (awk);
 *    qse_awk_seterrstr (awk, myerrstr);
 *    ...
 * }
 * @endcode
 */
void qse_awk_seterrstr (
	qse_awk_t*       awk,   /**< awk object */
	qse_awk_errstr_t errstr /**< error string getter */
);

/**
 * The qse_awk_geterrnum() function returns the number of the last error 
 * occurred.
 * @return error number
 */
qse_awk_errnum_t qse_awk_geterrnum (
	qse_awk_t* awk /**< awk object */
);

/**
 * The qse_awk_geterrlin() function returns the line number where the
 * last error has occurred.
 * @return line number
 */
qse_size_t qse_awk_geterrlin (
	qse_awk_t* awk /**< awk object */
);

/**
 * The qse_awk_geterrmsg() function returns the error message describing
 * the last error occurred.
 * @return error message
 */
const qse_char_t* qse_awk_geterrmsg (
	qse_awk_t* awk /**< awk object */
);

/**
 * The qse_awk_geterrinf() function copies error information into memory
 * pointed to by @a errinf from @a awk.
 */
void qse_awk_geterrinf (
	qse_awk_t*        awk,   /**< awk object */
	qse_awk_errinf_t* errinf /**< error information buffer */
);

/**
 * The qse_awk_seterrnum() function sets the error information omitting 
 * error location. You must pass a non-NULL for @a errarg if the specified
 * error number @a errnum requires one or more arguments to format an
 * error message.
 */
void qse_awk_seterrnum (
	qse_awk_t*        awk,    /**< awk object */
	qse_awk_errnum_t  errnum, /**< error number */
	const qse_cstr_t* errarg  /**< arguments to format error message */
);

void qse_awk_seterrinf (
	qse_awk_t*              awk, 
	const qse_awk_errinf_t* errinf
);

void qse_awk_geterror (
	qse_awk_t*         awk,
	qse_awk_errnum_t*  errnum, 
	const qse_char_t** errmsg,
	qse_size_t*        errlin
);

/**
 * The qse_awk_seterror() functon sets error information.
 */
void qse_awk_seterror (
	qse_awk_t*        awk,    /**< awk object */
	qse_awk_errnum_t  errnum, /**< error number */
	const qse_cstr_t* errarg, /**< argument array for formatting error message */
	qse_size_t        errlin  /**< line number */
);

int qse_awk_getoption (
	qse_awk_t* awk
);

void qse_awk_setoption (
	qse_awk_t* awk,
	int        opt
);

qse_size_t qse_awk_getmaxdepth (
	qse_awk_t* awk,
	int        type
);

void qse_awk_setmaxdepth (
	qse_awk_t* awk,
	int        types,
	qse_size_t depth
);

int qse_awk_getword (
	qse_awk_t*        awk, 
	const qse_cstr_t* okw,
	qse_cstr_t*       nkw
);

int qse_awk_unsetword (
	qse_awk_t*        awk,
	const qse_cstr_t* kw
);

void qse_awk_unsetallwords (
	qse_awk_t* awk
);

/**
 * The qse_awk_setword() function enables replacement of a name of a keyword,
 * intrinsic global variables, and intrinsic functions.
 *
 * If @a nkw is #QSE_NULL or @a nlen is zero and @a okw is #QSE_NULL or 
 * @a olen is zero, it unsets all word replacements; If @a nkw is #QSE_NULL or 
 * @a nlen is zero, it unsets the replacement for @a okw and @a olen; If 
 * all of them are valid, it sets the word replace for @a okw and @a olen 
 * to @a nkw and @a nlen.
 *
 * @return 0 on success, -1 on failure
 */
int qse_awk_setword (
	qse_awk_t*        awk,  /**< awk object */
	const qse_cstr_t* okw,  /**< old keyword */
	const qse_cstr_t* nkw   /**< new keyword */
);

/**
 * The qse_awk_addgbl() function adds an intrinsic global variable.
 * @return the ID of the global variable added on success, -1 on failure.
 */
int qse_awk_addgbl (
	qse_awk_t*        awk,   /**< awk object */
	const qse_char_t* name,  /**< a variable name */
	qse_size_t        len    /**< name length */
);

/**
 * The qse_awk_delgbl() function deletes an instrinsic global variable by name.
 * @return 0 on success, -1 on failure
 */
int qse_awk_delgbl (
	qse_awk_t*        awk,  /**< awk object */
	const qse_char_t* name, /**< a variable name */
	qse_size_t        len   /**< name length */
);

/**
 * The qse_awk_addfnc() function adds an intrinsic function.
 */
void* qse_awk_addfnc (
	qse_awk_t*        awk,
	const qse_char_t* name,
	qse_size_t        name_len, 
	int               when_valid,
	qse_size_t        min_args,
	qse_size_t        max_args, 
	const qse_char_t* arg_spec, 
	qse_awk_fnc_fun_t handler
);

/**
 * The qse_awk_delfnc() function deletes an intrinsic function by name.
 * @return 0 on success, -1 on failure
 */
int qse_awk_delfnc (
	qse_awk_t*        awk,  /**< awk object */
	const qse_char_t* name, /**< function name */
	qse_size_t        len   /**< name length */
);

/**
 * The qse_awk_clrfnc() function deletes all intrinsic functions
 */
void qse_awk_clrfnc (
	qse_awk_t* awk /**< awk object */
);

/**
 * The qse_awk_parse() function parses a source script, and optionally 
 * deparses it back. 
 *
 * It reads a source script by calling @a sio->in as shown in the pseudo code 
 * below:
 *
 * @code
 * n = sio->in (awk, QSE_AWK_SIO_OPEN);
 * if (n >= 0)
 * {
 *    while (n > 0)
 *       n = sio->in (awk, QSE_AWK_SIO_READ, buf, buf_size);
 *    sio->in (awk, QSE_AWK_SIO_CLOSE);
 * }
 * @endcode
 *
 * A negative number returned causes qse_awk_parse() to return failure;
 * 0 returned indicates the end of a stream; A positive number returned 
 * indicates successful opening of a stream or the length of the text read.
 *
 * If @a sio->out is not #QSE_NULL, it deparses the internal parse tree
 * composed of a source script and writes back the deparsing result by 
 * calling @a sio->out as shown below:
 *
 * @code
 * n = sio->out (awk, QSE_AWK_SIO_OPEN);
 * if (n >= 0)
 * {
 *    while (n > 0)
 *       n = sio->out (awk, QSE_AWK_SIO_WRITE, text, text_size);
 *    sio->out (awk, QSE_AWK_SIO_CLOSE);
 * }
 * @endcode
 * 
 * Unlike @a sf->in, the return value of 0 from @a sf->out is treated as
 * premature end of a stream; therefore, it causes qse_awk_parse() to return
 * failure.
 *
 * @return 0 on success, -1 on failure.
 */
int qse_awk_parse (
	qse_awk_t*     awk, /**< awk object */
	qse_awk_sio_t* sio  /**< source script I/O handler */
);

/**
 * The qse_awk_alloc() function allocates dynamic memory.
 * @return a pointer to a memory block on success, #QSE_NULL on failure
 */
void* qse_awk_alloc (
	qse_awk_t* awk,  /**< awk object */
	qse_size_t size  /**< size of memory to allocate in bytes */
);

/**
 * The qse_awk_realloc() function resizes a dynamic memory block.
 * @return a pointer to a memory block on success, #QSE_NULL on failure
 */
void* qse_awk_realloc (
	qse_awk_t* awk,  /**< awk object */
	void*      ptr,  /**< memory block */
	qse_size_t size  /**< new block size in bytes */
);


/**
 * The qse_awk_free() function frees dynamic memory allocated.
 */
void qse_awk_free (
	qse_awk_t* awk, /**< awk object */
	void*      ptr  /**< memory block to free */
);

/**
 * The qse_awk_strdup() function is used to duplicate a string using
 * the memory manager used by the associated qse_awk_t instance.
 * The new string should be freed using the qse_awk_free() function.
 *
 * @return a pointer to a new string duplicated of @a s on success, 
 *         #QSE_NULL on failure.
 */
qse_char_t* qse_awk_strdup (
	qse_awk_t*        awk, /**< awk object */
	const qse_char_t* str  /**< a string pointer */
);

/**
 * The qse_awk_strxdup() function is used to duplicate a string whose length
 * is as long as len characters using the memory manager used by the 
 * qse_awk_t instance. The new string should be freed using the qse_awk_free()
 * function.
 *
 * @return a pointer to a new string duplicated of @a s on success, 
 *         #QSE_NULL on failure.
 */
qse_char_t* qse_awk_strxdup (
	qse_awk_t*        awk, /**< awk object */
	const qse_char_t* str, /**< a string pointer */
	qse_size_t        len  /**< the number of character in a string */
);

qse_long_t qse_awk_strxtolong (
	qse_awk_t*         awk,
	const qse_char_t*  str,
	qse_size_t         len,
	int                base,
	const qse_char_t** endptr
);

qse_real_t qse_awk_strxtoreal (
	qse_awk_t*         awk,
	const qse_char_t*  str,
	qse_size_t         len, 
	const qse_char_t** endptr
);

qse_size_t qse_awk_longtostr (
	qse_awk_t*        awk,
	qse_long_t        value,
	int               radix,
	const qse_char_t* prefix,
	qse_char_t*       buf,
	qse_size_t        size
);

/**
 * The qse_awk_rtx_open() creates a runtime context.
 * @return a runtime context on success, #QSE_NULL on failure
 */
qse_awk_rtx_t* qse_awk_rtx_open (
	qse_awk_t*        awk, /**< awk object */
	qse_size_t        xtn, /**< size of extension in bytes */
	qse_awk_rio_t*    rio, /**< runtime IO handlers */
	const qse_cstr_t* arg  /**< arguments to set ARGV */
);

/**
 * The qse_awk_rtx_close() function destroys a runtime context.
 */
void qse_awk_rtx_close (
	qse_awk_rtx_t* rtx /**< runtime context */
);

/**
 * The qse_awk_rtx_loop() function executes the BEGIN block, pattern-action
 * blocks and the END blocks in an AWk program. It returns the global return 
 * value of which the reference count must be decremented when not necessary. 
 * Multiple invocations of the function for the lifetime of a runtime context 
 * is not desirable.
 *
 * The example shows typical usage of the function.
 * @code
 * rtx = qse_awk_rtx_open (awk, 0, rio, QSE_NULL);
 * if (rtx != QSE_NULL)
 * {
 *    retv = qse_awk_rtx_loop (rtx);
 *    if (retv != QSE_NULL) qse_awk_rtx_refdownval (rtx, retv);
 *    qse_awk_rtx_close (rtx);
 * }
 * @endcode
 *
 * @return return value on success, QSE_NULL on failure.
 */
qse_awk_val_t* qse_awk_rtx_loop (
	qse_awk_rtx_t* rtx /**< runtime context */
);

/**
 * The qse_awk_rtx_call() function invokes an AWK function. However, it is
 * not able to invoke an intrinsic function such as split(). 
 * The #QSE_AWK_PABLOCK option can be turned off to make illegal the BEGIN 
 * block, pattern-action blocks, and the END block.
 *
 * The example shows typical usage of the function.
 * @code
 * rtx = qse_awk_rtx_open (awk, 0, rio, QSE_NULL);
 * if (rtx != QSE_NULL)
 * {
 *     v = qse_awk_rtx_call (rtx, QSE_T("init"), QSE_NULL, 0);
 *     if (v != QSE_NULL) qse_awk_rtx_refdownval (rtx, v);
 *     qse_awk_rtx_call (rtx, QSE_T("fini"), QSE_NULL, 0);
 *     if (v != QSE_NULL) qse_awk_rtx_refdownval (rtx, v);
 *     qse_awk_rtx_close (rtx);
 * }
 * @endcode
 *
 * @return 0 on success, -1 on failure
 */
qse_awk_val_t* qse_awk_rtx_call (
	qse_awk_rtx_t*    rtx,  /**< runtime context */
	const qse_char_t* name, /**< function name */
	qse_awk_val_t**   args, /**< arguments to the function */
	qse_size_t        nargs /**< the number of arguments */
);

/**
 * The qse_awk_stopall() function aborts all active runtime contexts
 * invoked with the awk parameter.
 */
void qse_awk_stopall (
	qse_awk_t* awk /**< awk object */
);

/**
 * The qse_awk_shouldstop function tests if qse_awk_rtx_stop() has been called.
 */
qse_bool_t qse_awk_rtx_shouldstop (
	qse_awk_rtx_t* rtx
);

/**
 * The qse_awk_rtx_stop() function causes the active qse_awk_run() function to 
 * be aborted. 
 */
void qse_awk_rtx_stop (
	qse_awk_rtx_t* rtx
);

/**
 * The qse_awk_rtx_setrcb() function gets runtime callbacks.
 * @return #QSE_NULL if no callback is set. Otherwise, the pointer to a 
 *         callback set.
 * @sa qse_awk_rtx_setrcb
 */
qse_awk_rcb_t* qse_awk_rtx_getrcb (
	qse_awk_rtx_t* rtx /**< runtime context */
);

/**
 * The qse_awk_rtx_setrcb() function sets runtime callbacks.
 * @sa qse_awk_rtx_getrcb
 */
void qse_awk_rtx_setrcb (
	qse_awk_rtx_t* rtx, /**< runtime context */
	qse_awk_rcb_t* rcb  /**< callback set */
);

/**
 * The qse_awk_rtx_getnargs() gets the number of arguments passed to 
 * qse_awk_run().
 */
qse_size_t qse_awk_rtx_getnargs (
	qse_awk_rtx_t* rtx
);

/**
 * The qse_awk_rtx_getarg() function gets an argument passed to qse_awk_run().
 */
qse_awk_val_t* qse_awk_rtx_getarg (
	qse_awk_rtx_t* rtx,
	qse_size_t     idx
);

/**
 * The qse_awk_rtx_getgbl() gets the value of a global variable.
 * The global variable ID @a id is one of the predefined global 
 * variable IDs or a value returned by qse_awk_addgbl().
 * This function never fails so long as the ID is valid. Otherwise, 
 * you may get into trouble.
 *
 * @return a value pointer
 */
qse_awk_val_t* qse_awk_rtx_getgbl (
	qse_awk_rtx_t* rtx, /**< runtime context */
	int            id   /**< a global variable ID */
);

/**
 * The qse_awk_rtx_setgbl() sets the value of a global variable.
 */
int qse_awk_rtx_setgbl (
	qse_awk_rtx_t* rtx, 
	int            id,
	qse_awk_val_t* val
);

/**
 * The qse_awk_rtx_setretval() sets the return value of a function
 * when called from within a function handler. The caller doesn't
 * have to invoke qse_awk_rtx_refupval() and qse_awk_rtx_refdownval() 
 * with the value to be passed to qse_awk_rtx_setretval(). 
 * The qse_awk_rtx_setretval() will update its reference count properly
 * once the return value is set. 
 */
void qse_awk_rtx_setretval (
	qse_awk_rtx_t* rtx, /**< runtime context */
	qse_awk_val_t* val  /**< return value */
);

/**
 * The qse_awk_rtx_setfilename() function sets FILENAME.
 */
int qse_awk_rtx_setfilename (
	qse_awk_rtx_t*    rtx, /**< runtime context */
	const qse_char_t* str, /**< name pointer */
	qse_size_t        len  /**< name length */
);

/**
 * The qse_awk_rtx_setofilename() function sets OFILENAME.
 */
int qse_awk_rtx_setofilename (
	qse_awk_rtx_t*    rtx, /**< runtime context */
	const qse_char_t* str, /**< name pointer */
	qse_size_t        len  /**< name length */
);

/**
 * The qse_awk_rtx_getawk() function gets the owning awk object.
 */
qse_awk_t* qse_awk_rtx_getawk (
	qse_awk_rtx_t* rtx /**< runtime context */
);

/**
 * The qse_awk_rtx_getmmgr() function gets the memory manager of a runtime
 * context.
 */
qse_mmgr_t* qse_awk_rtx_getmmgr (
	qse_awk_rtx_t* rtx /**< runtime context */
);

/**
 * The qse_awk_rtx_getxtn() function gets the pointer to extension space.
 */
void* qse_awk_rtx_getxtn (
	qse_awk_rtx_t* rtx /**< runtime context */
);

/**
 * The qse_awk_rtx_getnvmap() gets the map of named variables 
 */
qse_map_t* qse_awk_rtx_getnvmap (
	qse_awk_rtx_t* rtx /**< runtime context */
);

/**
 * The qse_awk_rtx_geterrnum() function gets the number of the last error
 * occurred in a runtime context.
 * @return error number
 */
qse_awk_errnum_t qse_awk_rtx_geterrnum (
	qse_awk_rtx_t* rtx /**< runtime context */
);

qse_size_t qse_awk_rtx_geterrlin (
	qse_awk_rtx_t* rtx /**< runtime context */
);

const qse_char_t* qse_awk_rtx_geterrmsg (
	qse_awk_rtx_t* rtx /**< runtime context */
);

/**
 * The qse_awk_rtx_geterrinf() function copies error information into memory
 * pointed to by @a errinf from a runtime context @a rtx.
 */
void qse_awk_rtx_geterrinf (
	qse_awk_rtx_t*    rtx,   /**< runtime context */
	qse_awk_errinf_t* errinf /**< error information */
);

/**
 * The qse_awk_rtx_geterror() function retrieves error information from a 
 * runtime context @rtx. The error number is stored into memory pointed
 * to by @a errnum; the error message pointer into memory pointed to by 
 * @a errmsg; the error line into memory pointed to by @a errlin.
 */
void qse_awk_rtx_geterror (
	qse_awk_rtx_t*     rtx,    /**< runtime context */
	qse_awk_errnum_t*  errnum, /**< error number */
	const qse_char_t** errmsg, /**< error message */
	qse_size_t*        errlin  /**< error line */
);

void qse_awk_rtx_seterrnum (
	qse_awk_rtx_t*    rtx,    /**< runtime context */
	qse_awk_errnum_t  errnum, /**< error number */
	const qse_cstr_t* errarg  /**< arguments to format error message */
);

void qse_awk_rtx_seterrinf (
	qse_awk_rtx_t*          rtx,   /**< runtime context */
	const qse_awk_errinf_t* errinf /**< error information */
);

void qse_awk_rtx_seterror (
	qse_awk_rtx_t*    rtx,    /**< runtime context */
	qse_awk_errnum_t  errnum, /**< error number */
	const qse_cstr_t* errarg, /**< argument array for formatting error message */
	qse_size_t        errlin  /**< error line */
);

/**
 * The qse_awk_rtx_clrrec() function clears the input record ($0) 
 * and fields ($1 to $N).
 */
int qse_awk_rtx_clrrec (
	qse_awk_rtx_t* rtx,
	qse_bool_t     skip_inrec_line
);

int qse_awk_rtx_setrec (
	qse_awk_rtx_t*    rtx,
	qse_size_t        idx,
	const qse_char_t* str,
	qse_size_t        len
);

/* value manipulation functions */
qse_awk_val_t* qse_awk_rtx_makeintval (
	qse_awk_rtx_t* rtx,
	qse_long_t     v
);
qse_awk_val_t* qse_awk_rtx_makerealval (
	qse_awk_rtx_t* rtx,
	qse_real_t     v
);

qse_awk_val_t* qse_awk_rtx_makestrval0 (
	qse_awk_rtx_t*    rtx,
	const qse_char_t* str
);

qse_awk_val_t* qse_awk_rtx_makestrval (
	qse_awk_rtx_t*    rtx,
	const qse_char_t* str,
	qse_size_t        len
);

qse_awk_val_t* qse_awk_rtx_makestrval_nodup (
	qse_awk_rtx_t* rtx,
	qse_char_t*    str,
	qse_size_t     len
);

qse_awk_val_t* qse_awk_rtx_makestrval2 (
	qse_awk_rtx_t*    rtx,
	const qse_char_t* str1,
	qse_size_t        len1, 
	const qse_char_t* str2,
	qse_size_t        len2
);

qse_awk_val_t* qse_awk_rtx_makenstrval (
	qse_awk_rtx_t*    rtx,
	const qse_char_t* str,
	qse_size_t        len
);

qse_awk_val_t* qse_awk_rtx_makerexval (
	qse_awk_rtx_t*    rtx,
	const qse_char_t* buf,
	qse_size_t        len,
	void*             code
);

qse_awk_val_t* qse_awk_rtx_makemapval (
	qse_awk_rtx_t* rtx
);

qse_awk_val_t* qse_awk_rtx_makerefval (
	qse_awk_rtx_t*  rtx,
	int             id,
	qse_awk_val_t** adr
);

qse_bool_t qse_awk_rtx_isstaticval (
	qse_awk_rtx_t* rtx,
	qse_awk_val_t* val
);

/**
 * The qse_awk_rtx_refupval() function increments a reference count of a 
 * value @a val.
 */
void qse_awk_rtx_refupval (
	qse_awk_rtx_t* rtx, /**< runtime context */
	qse_awk_val_t* val  /**< a value */
);

/**
 * The qse_awk_rtx_refdownval() function decrements a reference count of
 * a value @a val. It destroys the value if it has reached the count of 0.
 */
void qse_awk_rtx_refdownval (
	qse_awk_rtx_t* rtx, /**< runtime context */
	qse_awk_val_t* val  /**< value pointer */
);

/**
 * The qse_awk_rtx_refdownval() function decrements a reference count of
 * a value @a val. It does not destroy the value if it has reached the 
 * count of 0.
 */
void qse_awk_rtx_refdownval_nofree (
	qse_awk_rtx_t* rtx, /**< runtime context */
	qse_awk_val_t* val  /**< value pointer */
);

/**
 * The qse_awk_rtx_valtobool() function converts a value @a val to a boolean
 * value.
 */
qse_bool_t qse_awk_rtx_valtobool (
	qse_awk_rtx_t* rtx, /**< runtime context */
	qse_awk_val_t* val  /**< value pointer */
);

/**
 * The qse_awk_rtx_valtostr() function converts a value @a val to a string as 
 * instructed in the parameter out. Before the call to the function, you 
 * should initialize a variable of the qse_awk_rtx_valtostr_out_t type.
 *
 * The type field is one of the following qse_awk_rtx_valtostr_type_t values:
 *
 * - QSE_AWK_RTX_VALTOSTR_CPL
 * - QSE_AWK_RTX_VALTOSTR_CPLDUP
 * - QSE_AWK_RTX_VALTOSTR_STRP
 * - QSE_AWK_RTX_VALTOSTR_STRPCAT
 *
 * It can optionally be ORed with QSE_AWK_RTX_VALTOSTR_PRINT. The option
 * causes the function to use OFMT for real number conversion. Otherwise,
 * it uses CONVFMT. 
 *
 * You should initialize or free other fields before and after the call 
 * depending on the type field as shown below:
 *  
 * If you have a static buffer, use QSE_AWK_RTX_VALTOSTR_CPL.
 * @code
 * qse_awk_rtx_valtostr_out_t out;
 * qse_char_t buf[100];
 * out.type = QSE_AWK_RTX_VALTOSTR_CPL;
 * out.u.cpl.ptr = buf;
 * out.u.cpl.len = QSE_COUNTOF(buf);
 * if (qse_awk_rtx_valtostr (rtx, v, &out) == QSE_NULL)  goto oops;
 * qse_printf (QSE_T("%.*s\n"), (int)out.u.cpl.len, out.u.cpl.ptr);
 * @endcode
 * 
 * When unsure of the size of the string after conversion, you can use
 * QSE_AWK_RTX_VALTOSTR_CPLDUP. However, you should free the memory block
 * pointed to by the u.cpldup.ptr field after use.
 * @code
 * qse_awk_rtx_valtostr_out_t out;
 * out.type = QSE_AWK_RTX_VALTOSTR_CPLDUP;
 * if (qse_awk_rtx_valtostr (rtx, v, &out) == QSE_NULL)  goto oops;
 * qse_printf (QSE_T("%.*s\n"), (int)out.u.cpldup.len, out.u.cpldup.ptr);
 * qse_awk_rtx_free (rtx, out.u.cpldup.ptr);
 * @endcode
 *
 * You may like to store the result in a dynamically resizable string.
 * Consider QSE_AWK_RTX_VALTOSTR_STRP.
 * @code
 * qse_awk_rtx_valtostr_out_t out;
 * qse_str_t str;
 * qse_str_init (&str, qse_awk_rtx_getmmgr(rtx), 100);
 * out.type = QSE_AWK_RTX_VALTOSTR_STRP;
 * out.u.strp = str;
 * if (qse_awk_rtx_valtostr (rtx, v, &out) == QSE_NULL)  goto oops;
 * qse_printf (QSE_T("%.*s\n"), 
 *     (int)QSE_STR_LEN(out.u.strp), QSE_STR_PTR(out.u.strp));
 * qse_str_fini (&str);
 * @endcode
 * 
 * If you want to append the converted string to an existing dynamically 
 * resizable string, QSE_AWK_RTX_VALTOSTR_STRPCAT is the answer. The usage is
 * the same as QSE_AWK_RTX_VALTOSTR_STRP except that you have to use the 
 * u.strpcat field instead of the u.strp field.
 *
 * In the context where @a val is determined to be of the type
 * #QSE_AWK_VAL_STR, you may access its string pointer and length directly
 * instead of calling this function.
 *
 * @return the pointer to a string converted on success, #QSE_NULL on failure
 */
qse_char_t* qse_awk_rtx_valtostr (
	qse_awk_rtx_t*              rtx, /**< runtime context */
	qse_awk_val_t*              val, /**< value to convert */
	qse_awk_rtx_valtostr_out_t* out  /**< output buffer */
);

/**
 * The qse_awk_rtx_valtocpldup() function provides a shortcut to the 
 * qse_awk_rtx_valtostr() function with the QSE_AWK_RTX_VALTOSTR_CPLDUP type.
 * It returns the pointer to a string converted from @a val and stores its 
 * length to memory pointed to by @a len. You should free the returned
 * memory block after use. See the code snippet below for a simple usage.
 *
 * @code
 * ptr = qse_awk_rtx_valtocpldup (rtx, v, &len);
 * if (str == QSE_NULL) handle_error();
 * qse_printf (QSE_T("%.*s\n"), (int)len, ptr);
 * qse_awk_rtx_free (rtx, ptr);
 * @endcode
 *
 * @return the pointer to a string converted on success, #QSE_NULL on failure
 */
qse_char_t* qse_awk_rtx_valtocpldup (
	qse_awk_rtx_t* rtx, /**< runtime context */
	qse_awk_val_t* val, /**< value to convert */
	qse_size_t*    len  /**< result length */
);

/**
 * The qse_awk_rtx_valtonum() function converts a value to a number. 
 * If the value is converted to a long number, it is stored in the memory
 * pointed to by l and 0 is returned. If the value is converted to a real 
 * number, it is stored in the memory pointed to by r and 1 is returned.
 * The function never fails as long as @a val points to a valid value.
 *
 * The code below shows how to convert a value to a number and determine
 * if it is an integer or a floating-point number.
 *
 * @code
 * qse_long_t l;
 * qse_real_t r;
 * int n;
 * n = qse_awk_rtx_valtonum (v, &l, &r);
 * if (n <= -1) error ();
 * else if (n == 0) print_long (l);
 * else if (n >= 1) print_real (r);
 * @endcode
 *
 * @return -1 on failure, 0 if converted to a long number, 1 if converted to 
 *         a floating-point number.
 */
int qse_awk_rtx_valtonum (
	qse_awk_rtx_t* rtx,
	qse_awk_val_t* val,
	qse_long_t*    l,
	qse_real_t*    r
);

/**
 * The qse_awk_rtx_strtonum() function converts a string to a number.
 * A numeric string in the valid decimal, hexadecimal(0x), binary(0b), 
 * octal(0) notation is converted to an integer and it is stored into
 * memory pointed to by @a l; A string containng '.', 'E', or 'e' is 
 * converted to a floating-pointer number and it is stored into memory
 * pointed to by @a r. If @a strict is 0, the function takes up to the last
 * valid character and never fails. If @a strict is non-zero, an invalid 
 * character causes the function to return an error.
 *
 * @return 0 if converted to an integer,
 *         1 if converted to a floating-point number
 *         -1 on error.
 */
int qse_awk_rtx_strtonum (
	qse_awk_rtx_t*    rtx, /**< runtime context */
	int               strict, /**< determines to perform strict check */
	const qse_char_t* ptr, /**< points to a string to convert */
	qse_size_t        len, /**< number of characters in a string */
	qse_long_t*       l,   /**< stores a converted integer */
	qse_real_t*       r    /**< stores a converted floating-poing number */
);

/**
 * The qse_awk_rtx_alloc() function allocats a memory block of @a size bytes
 * using the memory manager associated with a runtime context @a rtx. 
 * @return the pointer to a memory block on success, #QSE_NULL on failure.
 */
void* qse_awk_rtx_alloc (
	qse_awk_rtx_t* rtx, /**< runtime context */
	qse_size_t     size /**< block size in bytes */
);

/**
 * The qse_awk_rtx_realloc() function resizes a memory block pointed to
 * by @a ptr to @a size bytes using the memory manager associated with 
 * a runtime context @a rtx. 
 * @return the pointer to a memory block on success, #QSE_NULL on failure.
 */
void* qse_awk_rtx_realloc (
	qse_awk_rtx_t* rtx, /**< runtime context */
	void*          ptr, /**< memory block */
	qse_size_t     size /**< block size in bytes */
);

/**
 * The qse_awk_rtx_free() function frees a memory block pointed to by @a ptr
 * using the memory manager of a runtime ocntext @a rtx.
 */
void qse_awk_rtx_free (
	qse_awk_rtx_t* rtx, /**< runtime context */
	void*          ptr  /**< memory block pointer */
);

#ifdef __cplusplus
}
#endif

#endif
